from config import get_config

import collections
import json
import os
import subprocess
import shutil
import tempfile
import time
import urlparse
import luigi
import luigi.s3
import tdclient

import logging
logger = logging.getLogger('luigi-interface')

class BulkImportUploadContext(object):
    def __init__(self, env, log_dir='log', tmp_dir='tmp'):
        self.env = env
        self.cur_dir = os.getcwd()
        self.cur_env = os.environ.copy()
        self.log_dir = log_dir
        self.tmp_dir = tmp_dir

    def __enter__(self):
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)
        if not os.path.exists(self.tmp_dir):
            os.makedirs(self.tmp_dir)
        os.chdir(self.log_dir)
        os.environ.update(self.env)
        return self

    def __exit__(self, type, value, tb):
        os.environ.clear()
        os.environ.update(self.cur_env)
        os.chdir(self.cur_dir)
        shutil.rmtree(self.tmp_dir)

class BulkImportTarget(luigi.Target):
    config = get_config()

    def __init__(self, session_id):
        self._session_id = session_id
        self._session = None

    def session(self):
        if not self._session:
            try:
                client = self.config.get_client()
                self._session = client.bulk_import(self._session_id)
            except tdclient.api.NotFoundError:
                self._session = None
        return self._session

    def exists(self):
        return self.session() != None

class BulkImportSession(luigi.Task):
    config = get_config()
    session = luigi.Parameter()
    options = luigi.Parameter(significant=False)

    def output(self):
        return BulkImportTarget(self.session)

    def run(self):
        options = json.loads(self.options)
        client = self.config.get_client()
        client.create_bulk_import(self.session, options['database'], options['table'])

class BulkImportUpload(luigi.Task):
    config = get_config()
    session = luigi.Parameter()
    options = luigi.Parameter(significant=False)

    def requires(self):
        return BulkImportSession(self.session, options=self.options)

    def complete(self):
        client = self.config.get_client()
        try:
            session = client.bulk_import(self.session)
        except tdclient.api.NotFoundError:
            return False
        else:
            return session.upload_frozen

    @property
    def log_dir(self):
        luigi_config = luigi.configuration.get_config()
        return os.path.abspath(os.path.join(luigi_config.get('td', 'log-dir', 'log'), self.session))

    @property
    def tmp_dir(self):
        luigi_config = luigi.configuration.get_config()
        return os.path.abspath(os.path.join(luigi_config.get('td', 'tmp-dir', 'tmp'), self.session))

    def full_path(self):
        options = json.loads(self.options)
        return [options['path']]

    def run_upload(self):
        options = json.loads(self.options)
        args = [
            'td',
            '-e', self.config.endpoint,
            '-k', self.config.apikey,
            'import:upload',
            self.session,
            '--format', 'csv',
            '--output', os.path.join(self.tmp_dir, 'parts'),
            '--error-records-output', os.path.join(self.log_dir, 'error-records'),
            '--time-column', options['time_column'],
            '--time-format', options['time_format'],
            '--encoding', 'utf-8',
            '--prepare-parallel', '8',
            '--parallel', '8',
            '--columns', ','.join([c.split(':')[0] for c in options['columns']]),
            '--column-types', ','.join([c.split(':')[1] for c in options['columns']]),
            '--error-records-handling', 'abort',
            '--empty-as-null-if-numeric',
        ]
        if options['column_header']:
            args += ['--column-header']
        args += self.full_path()
        env = {}
        env['TD_TOOLBELT_JAR_UPDATE'] = '0'
        if options['default_timezone']:
            env['TZ'] = options['default_timezone']
        with BulkImportUploadContext(env, log_dir=self.log_dir, tmp_dir=self.tmp_dir):
            logger.debug(' '.join(args))
            p = subprocess.Popen(args, stderr=subprocess.PIPE)
            out, err = p.communicate()
            if p.returncode != 0:
                raise RuntimeError(err)

    def run(self):
        self.run_upload()
        client = self.config.get_client()
        client.freeze_bulk_import(self.session)

class BulkImportPerform(luigi.Task):
    config = get_config()
    session = luigi.Parameter()
    options = luigi.Parameter(significant=False)

    def requires(self):
        return BulkImportUpload(self.session, options=self.options)

    def complete(self):
        client = self.config.get_client()
        try:
            session = client.bulk_import(self.session)
        except tdclient.api.NotFoundError:
            return False
        else:
            return session.status == 'ready'

    def run(self):
        client = self.config.get_client()
        session = client.bulk_import(self.session)
        if session.status == 'uploading':
            job = client.perform_bulk_import(self.session)
        else:
            job = client.job(session.job_id)
        logger.info("%s: performing...", self)
        while not job.finished():
            time.sleep(2)

class BulkImport(luigi.Task):
    config = get_config()
    session = luigi.Parameter()
    database = luigi.Parameter(significant=False)
    table = luigi.Parameter(significant=False)
    format = luigi.Parameter(significant=False, default='csv')
    columns = luigi.Parameter(significant=False)
    time_column = luigi.Parameter(significant=False)
    time_format = luigi.Parameter(significant=False)
    default_timezone = luigi.Parameter(significant=False)
    # CSV/TSV options
    column_header = luigi.BooleanParameter(significant=False, default=False)

    def complete(self):
        client = self.config.get_client()
        try:
            session = client.bulk_import(self.session)
        except tdclient.api.NotFoundError:
            return False
        else:
            return session.status == 'committed'

    def get_path(self):
        target = self.input()
        # LocalTarget
        if isinstance(target, luigi.LocalTarget):
            return os.path.abspath(target.path)
        # S3Target
        if isinstance(target, luigi.s3.S3Target):
            url = urlparse.urlparse(target.path)
            luigi_config = luigi.configuration.get_config()
            return "s3://{aws_access_key_id}:{aws_secret_access_key}@/{bucket}{path}".format(
                aws_access_key_id = luigi_config.get('s3', 'aws_access_key_id', os.environ.get('AWS_ACCESS_KEY_ID')),
                aws_secret_access_key = luigi_config.get('s3', 'aws_secret_access_key', os.environ.get('AWS_SECRET_ACCESS_KEY')),
                bucket = url.hostname,
                path = url.path
            )
        raise ValueError('unsupported target: {0}'.format(target))

    def run(self):
        options = {
            'database': self.database,
            'table': self.table,
            'path': self.get_path(),
            'format': self.format,
            'columns': self.columns,
            'time_column': self.time_column,
            'time_format': self.time_format,
            'default_timezone': self.default_timezone,
            'column_header': self.column_header,
        }

        # upload and perform
        yield BulkImportPerform(self.session, options=json.dumps(options))

        # commit
        client = self.config.get_client()
        client.commit_bulk_import(self.session)



from client import ResultProxy

import os
import shutil
import tempfile
import tdclient

class MockDatabase(object):
    def __init__(self, name):
        self.name = name

class MockTable(object):
    def __init__(self, database, name):
        self.database = database
        self.name = name
        self.schema = []

class MockJob(object):
    def __init__(self, spec):
        self.spec = spec
        self.job_id = spec['job_id']
        self.url = 'https://mock.example.com/jobs/{0}'.format(spec['job_id'])
        self._result_size = spec.get('size')
        self._hive_result_schema = spec.get('description', [])

    def _update_status(self):
        pass

    def kill(self):
        pass

    def status(self):
        return self.spec['status']

    def success(self):
        return self.status() == 'success'

    def finished(self):
        return self.status() != 'pending'

    def result(self):
        return iter(self.spec['rows'])

class MockClient(object):
    def __init__(self, databases, tables, jobs):
        self._databases = databases
        self._tables = tables
        self._jobs = jobs

    def create_database(self, name):
        self._databases.append(name)

    def databases(self):
        return [MockDatabase(name) for name in self._databases]

    def create_log_table(self, database, name):
        self._tables.append("{0}.{1}".format(database, name))

    def update_schema(self, database, name, schema):
        pass

    def table(self, database, name):
        if "{0}.{1}".format(database, name) in self._tables:
            return MockTable(database, name)
        raise tdclient.api.NotFoundError(database, name)

    def jobs(self):
        return [MockJob(job) for job in self._jobs]

    def job(self, job_id):
        for job in self.jobs():
            if job.job_id == job_id:
                return job

    def query(self, database, query, type='hive', result_url=None):
        return MockJob(self._jobs[0])

class TestConfig(object):
    def __init__(self, databases=[], tables=[], jobs=[]):
        self._databases = databases
        self._tables = tables
        self._jobs = jobs
        self.tmp_dir = None

    def setUp(self):
        if not self.tmp_dir:
            self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        if self.tmp_dir:
            shutil.rmtree(self.tmp_dir)
            self.tmp_dir = None

    def get_tmp_path(self, filename):
        return os.path.join(self.tmp_dir, filename)

    def get_client(self):
        return MockClient(databases=self._databases, tables=self._tables, jobs=self._jobs)



from client import ResultProxy
from test_helper import MockJob

from unittest import TestCase
from nose.tools import eq_, raises

import tempfile

SUCCESS_JOB = {
    'job_id': 1,
    'status': 'success',
    'size': 20,
    'description': [['cnt', 'int']],
    'rows': [[5000]],
}

class ResultProxyTestCase(TestCase):
    def test_with_successful_job(self):
        result = ResultProxy(MockJob(SUCCESS_JOB))
        eq_(result.job_id, SUCCESS_JOB['job_id'])
        eq_(result.status, SUCCESS_JOB['status'])
        eq_(result.size, SUCCESS_JOB['size'])
        eq_(result.description, SUCCESS_JOB['description'])
        eq_(list(result), SUCCESS_JOB['rows'])
        with tempfile.NamedTemporaryFile() as f:
            result.to_csv(f.name)
            eq_(f.read(), "cnt\n5000\n")



from luigi_td.targets.result import ResultTarget

import urllib

import logging
logger = logging.getLogger('luigi-interface')

__all__ = ['TableauServerResultTarget', 'TableauOnlineResultTarget']

class TableauServerResultTarget(ResultTarget):
    # required
    server = None
    username = None
    password = None
    datasource = None
    # optional
    ssl = 'true'
    ssl_verify = 'true'
    version = None
    site = None
    project = None
    mode = 'replace'

    def get_result_url(self):
        reqs = {}
        for name in ['server', 'username', 'password', 'datasource']:
            if getattr(self, name) is None:
                raise TypeError('missing option "{0}" for {1}'.format(name, self))
            reqs[name] = urllib.quote(getattr(self, name))
        params = {
            'ssl': self.ssl,
            'ssl_verify': self.ssl_verify,
            'version': self.version,
            'site': self.site,
            'project': self.project,
            'mode': self.mode,
        }
        reqs['params'] = urllib.urlencode([(key, params[key]) for key in params if params[key] is not None])
        return "tableau://{username}:{password}@{server}/{datasource}?{params}".format(**reqs)

class TableauOnlineResultTarget(TableauServerResultTarget):
    server = 'online.tableausoftware.com'
    version = 'online'



from luigi_td.config import get_config

import luigi
import tdclient

import logging
logger = logging.getLogger('luigi-interface')

__all__ = ['DatabaseTarget', 'TableTarget', 'SchemaError']

class SchemaError(Exception):
    pass

class DatabaseTarget(luigi.Target):
    def __init__(self, database_name, config=None):
        self.database_name = database_name
        self.config = config or get_config()

    def exists(self):
        client = self.config.get_client()
        return self.database_name in [db.name for db in client.databases()]

class TableTarget(luigi.Target):
    def __init__(self, database_name, table_name, schema=None, empty=False, config=None):
        self.database_name = database_name
        self.table_name = table_name
        self.schema = schema or []
        self.empty = empty
        self.config = config or get_config()

    def exists(self):
        client = self.config.get_client()
        try:
            t = client.table(self.database_name, self.table_name)
        except tdclient.api.NotFoundError:
            return False
        if self.empty:
            if t and t.count == 0 and [':'.join(c) for c in t.schema] == list(self.schema):
                return True
            else:
                logger.debug('Deleting table: {0}.{1}'.format(self.database_name, self.table_name))
                client.delete_table(self.database_name, self.table_name)
                return False
        else:
            if t:
                table_schema = [str(':'.join(c)) for c in t.schema]
                if table_schema != list(self.schema):
                    logger.error('Current schema: {0}'.format(table_schema))
                    logger.error('Expected schema: {0}'.format(self.schema))
                    raise SchemaError('table schema for {0}.{1} does not match'.format(self.database_name, self.table_name))
                return True
            else:
                return False



from ..test_helper import TestConfig
from ..client import ResultProxy
from result import ResultTarget

from unittest import TestCase
from nose.tools import eq_, raises

test_config = TestConfig(
    jobs = [
        {
            'job_id': 1,
            'status': 'success',
            'size': 20,
            'description': [['cnt', 'int']],
            'rows': [[5000]],
        },
        {
            'job_id': 2,
            'status': 'error',
            'size': 20,
        },
        {
            'job_id': 3,
            'status': 'killed',
            'size': 20,
        },
        {
            'job_id': 4,
            'status': 'queued',
        },
        {
            'job_id': 5,
            'status': 'running',
        },
    ]
)

class ResultTargetTestCase(TestCase):
    def setUp(self):
        test_config.setUp()
        for job in test_config.get_client().jobs():
            result = ResultProxy(job)
            target = ResultTarget(test_config.get_tmp_path('{0}.job'.format(result.status)),
                                  config=test_config)
            target.save_result_state(result)

    def tearDown(self):
        test_config.tearDown()

    def test_not_exists_when_state_file_does_not_found(self):
        target = ResultTarget(test_config.get_tmp_path('dummy.job'), config=test_config)
        eq_(target.exists(), False)

    def test_not_exists_when_status_is_not_success(self):
        for job in test_config.get_client().jobs():
            status = job.status()
            if status != 'success':
                target = ResultTarget(test_config.get_tmp_path('{0}.job'.format(status)),
                                      config=test_config)
                eq_(target.exists(), False)

    def test_exists_when_status_is_success(self):
        target = ResultTarget(test_config.get_tmp_path('success.job'), config=test_config)
        eq_(target.exists(), True)
        eq_(type(target.result), ResultProxy)



from ..test_helper import TestConfig
from td import DatabaseTarget, TableTarget, SchemaError

from unittest import TestCase
from nose.tools import eq_, raises

import urllib
import urlparse

test_config = TestConfig(
    databases = ['test_db'],
    tables = ['test_db.test_table'],
)

class DatabaseTargetTestCase(TestCase):
    def test_exists(self):
        target = DatabaseTarget('test_db', config=test_config)
        eq_(target.exists(), True)

    def test_not_exists(self):
        target = DatabaseTarget('dummy_db', config=test_config)
        eq_(target.exists(), False)

class TableTargetTestCase(TestCase):
    def test_exists(self):
        target = TableTarget('test_db', 'test_table', config=test_config)
        eq_(target.exists(), True)

    def test_table_not_exists(self):
        target = TableTarget('test_db', 'dummy_table', config=test_config)
        eq_(target.exists(), False)

    def test_database_not_exists(self):
        target = TableTarget('dummy_db', 'test_table', config=test_config)
        eq_(target.exists(), False)

    @raises(SchemaError)
    def test_schema_unmatch(self):
        target = TableTarget('test_db', 'test_table', schema=[['c', 'int']], config=test_config)
        target.exists()
